#ifndef __SPXCONFIG_H__
#define __SPXCONFIG_H__

#define CMAKE_BUILD_TYPE "None"

#define SOPLEX_VERSION_MAJOR 6
#define SOPLEX_VERSION_MINOR 0
#define SOPLEX_VERSION_PATCH 4
#define SOPLEX_VERSION_SUB 0
#define SOPLEX_VERSION_API 11
#define SOPLEX_WITH_BOOST
/* #undef SOPLEX_WITH_CPPMPF */
/* #undef SOPLEX_WITH_FLOAT128 */
#define SOPLEX_WITH_GMP
#define SOPLEX_WITH_MPFR
/* #undef SOPLEX_WITH_PAPILO */
#define SOPLEX_WITH_ZLIB

#endif
